from unittest.mock import Mock

from travertino.node import Node

from toga.style.applicator import TogaApplicator


class ExampleNode(Node):
    def __init__(self, name, style, size=None, children=None):
        super().__init__(
            style=style, children=children, applicator=TogaApplicator(self)
        )

        self.name = name
        self._impl = Mock()
        if size:
            self.intrinsic.width = size[0]
            self.intrinsic.height = size[1]

        self.refresh = Mock()

    def __repr__(self):
        return f"<{self.name}>"

    def __html__(self, depth=0):
        """Debugging helper - output the HTML interpretation of this layout"""
        if depth:
            tag = "div"
        else:
            tag = "body"
        lines = []

        # Add an interpretation of intrinsic width
        extra_style = [""]
        if self.intrinsic.width:
            try:
                extra_style.append(f"min-width: {self.intrinsic.width.value}px;")
            except AttributeError:
                extra_style.append(f"min-width: {self.intrinsic.width}px;")
        if self.intrinsic.height:
            try:
                extra_style.append(f"min-height: {self.intrinsic.height.value}px;")
            except AttributeError:
                extra_style.append(f"min-height: {self.intrinsic.height}px;")
        extra_css = " ".join(extra_style)

        lines.append(
            "    " * (depth + 1)
            + f'<{tag} id="{self.name}" style="{self.style.__css__()}{extra_css}">'
        )
        if self.children:
            for child in self.children:
                lines.append(child.__html__(depth=depth + 1))
            lines.append("    " * (depth + 1) + f"</{tag}>")
        else:
            lines[-1] = lines[-1] + f"</{tag}>"
        return "\n".join(lines)

    def refresh(self):
        # We're directly modifying styles and computing layouts for specific
        # viewports, so we don't need to trigger layout changes when a style is
        # changed.
        pass


class ExampleViewport:
    def __init__(self, width, height):
        self.height = height
        self.width = width


def _assert_layout(node, expected_layout):
    assert (
        node.layout.absolute_content_left,
        node.layout.absolute_content_top,
    ) == expected_layout["origin"], (
        f"origin of {node} ({node.layout.absolute_content_left}, {node.layout.absolute_content_top}) "
        f"doesn't match expected {expected_layout['origin']}"
    )
    assert (node.layout.content_width, node.layout.content_height) == expected_layout[
        "content"
    ], (
        f"content size of {node} ({node.layout.content_width}, {node.layout.content_height}) "
        f"doesn't match expected {expected_layout['content']}"
    )

    n_children = len(node.children)
    expected_n_children = len(expected_layout.get("children", []))
    assert n_children == expected_n_children, (
        f"number of children of {node} ({n_children}) "
        f"doesn't match expected {expected_n_children}"
    )

    for child, sublayout in zip(node.children, expected_layout.get("children", [])):
        _assert_layout(child, sublayout)


def assert_layout(node, min_size, expected_size, expected_layout):
    """Assert that a node's full recursive layout matches expectation."""
    assert (
        node.layout.min_width,
        node.layout.min_height,
    ) == min_size, (
        f"minimum size {node.layout.min_width}x{node.layout.min_height} "
        f"doesn't match expected {min_size[0]}x{min_size[1]}"
    )

    assert (
        node.layout.width,
        node.layout.height,
    ) == expected_size, (
        f"final size {node.layout.width}x{node.layout.height} "
        f"doesn't match expected {expected_size[0]}x{expected_size[1]}"
    )
    _assert_layout(node, expected_layout)



from __future__ import annotations

import warnings
from collections.abc import Iterable
from typing import Any, Protocol, TypeVar

import toga
from toga.handlers import wrapped_handler
from toga.sources import ListSource, Source

from .base import StyleT, Widget

SourceT = TypeVar("SourceT", bound=Source)


class OnChangeHandler(Protocol):
    def __call__(self, widget: Selection, /, **kwargs: Any) -> object:
        """A handler to invoke when the value is changed.

        :param widget: The Selection that was changed.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class Selection(Widget):
    def __init__(
        self,
        id: str | None = None,
        style: StyleT | None = None,
        items: SourceT | Iterable | None = None,
        accessor: str | None = None,
        value: object | None = None,
        on_change: toga.widgets.selection.OnChangeHandler | None = None,
        enabled: bool = True,
        on_select: None = None,  # DEPRECATED
    ):
        """Create a new Selection widget.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style will be
            applied to the widget.
        :param items: Initial :any:`items` to display for selection.
        :param accessor: The accessor to use to extract display values from the list of
            items. See :any:`items` and :any:`value` for details on how
            ``accessor`` alters the interpretation of data in the Selection.
        :param value: Initial value for the selection. If unspecified, the first item in
            ``items`` will be selected.
        :param on_change: Initial :any:`on_change` handler.
        :param enabled: Whether the user can interact with the widget.
        """
        super().__init__(id=id, style=style)

        ######################################################################
        # 2023-05: Backwards compatibility
        ######################################################################
        if on_select:  # pragma: no cover
            if on_change:
                raise ValueError("Cannot specify both on_select and on_change")
            else:
                warnings.warn(
                    "Selection.on_select has been renamed Selection.on_change",
                    DeprecationWarning,
                )
                on_change = on_select
        ######################################################################
        # End backwards compatibility.
        ######################################################################

        self._items: SourceT | ListSource

        self.on_change = None  # needed for _impl initialization
        self._impl = self.factory.Selection(interface=self)

        self._accessor = accessor
        self.items = items
        if value:
            self.value = value

        self.on_change = on_change
        self.enabled = enabled

    @property
    def items(self) -> SourceT | ListSource:
        """The items to display in the selection.

        When setting this property:

        * A :any:`Source` will be used as-is. It must either be a :any:`ListSource`, or
          a custom class that provides the same methods.

        * A value of None is turned into an empty ListSource.

        * Otherwise, the value must be an iterable, which is copied into a new
          ListSource using the widget's accessor, or "value" if no accessor was
          specified. Items are converted as shown :ref:`here <listsource-item>`.
        """
        return self._items

    @items.setter
    def items(self, items: SourceT | Iterable | None) -> None:
        if self._accessor is None:
            accessors = ["value"]
        else:
            accessors = [self._accessor]

        if items is None:
            self._items = ListSource(accessors=accessors, data=[])
        elif isinstance(items, Source):
            if self._accessor is None:
                raise ValueError("Must specify an accessor to use a data source")
            self._items = items
        else:
            self._items = ListSource(accessors=accessors, data=items)

        self._items.add_listener(self._impl)

        # Temporarily halt notifications
        orig_on_change = self._on_change
        self.on_change = None

        # Clear the widget, and insert all the data rows
        self._impl.clear()
        for index, item in enumerate(self.items):
            self._impl.insert(index, item)

        # Restore the original change handler and trigger it.
        self._on_change = orig_on_change
        self.on_change()

        self.refresh()

    def _title_for_item(self, item: Any) -> str:
        """Internal utility method; return the display title for an item"""
        if self._accessor:
            title = getattr(item, self._accessor)
        else:
            title = item.value

        return str(title).split("\n")[0]

    @property
    def value(self) -> object | None:
        """The currently selected item.

        Returns None if there are no items in the selection.

        If an ``accessor`` was specified when the Selection was constructed, the value
        returned will be Row objects from the ListSource; to change the selection, a Row
        object from the ListSource must be provided.

        If no ``accessor`` was specified when the Selection was constructed, the value
        returned will be the value stored as the ``value`` attribute on the Row object.
        When setting the value, the widget will search for the first Row object whose
        ``value`` attribute matches the provided value. In practice, this means that you
        can treat the selection as containing a list of literal values, rather than a
        ListSource containing Row objects.
        """
        index = self._impl.get_selected_index()
        if index is None:
            return None

        item = self._items[index]
        # If there was no accessor specified, the data values are literals.
        # Dereference the value out of the Row object.
        if item and self._accessor is None:
            return item.value
        return item

    @value.setter
    def value(self, value: object) -> None:
        try:
            if self._accessor is None:
                item = self._items.find(dict(value=value))
            else:
                item = value

            index = self._items.index(item)
            self._impl.select_item(index=index, item=item)
        except ValueError:
            raise ValueError(f"{value!r} is not a current item in the selection")

    @property
    def on_change(self) -> OnChangeHandler:
        """Handler to invoke when the value of the selection is changed, either by the user
        or programmatically."""
        return self._on_change

    @on_change.setter
    def on_change(self, handler: toga.widgets.selection.OnChangeHandler) -> None:
        self._on_change = wrapped_handler(self, handler)

    ######################################################################
    # 2023-05: Backwards compatibility
    ######################################################################

    @property
    def on_select(self) -> OnChangeHandler:
        """**DEPRECATED**: Use ``on_change``"""
        warnings.warn(
            "Selection.on_select has been renamed Selection.on_change.",
            DeprecationWarning,
        )
        return self.on_change

    @on_select.setter
    def on_select(self, handler: toga.widgets.selection.OnChangeHandler) -> None:
        warnings.warn(
            "Selection.on_select has been renamed Selection.on_change.",
            DeprecationWarning,
        )
        self.on_change = handler



from __future__ import annotations

from typing import Literal

from toga.constants import Direction

from .base import StyleT, Widget


class Divider(Widget):
    HORIZONTAL = Direction.HORIZONTAL
    VERTICAL = Direction.VERTICAL

    def __init__(
        self,
        id: str | None = None,
        style: StyleT | None = None,
        direction: Direction = HORIZONTAL,
    ):
        """Create a new divider line.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style will be
            applied to the widget.
        :param direction: The direction in which the divider will be drawn. Either
            :attr:`~toga.constants.Direction.HORIZONTAL` or
            :attr:`~toga.constants.Direction.VERTICAL`; defaults to
            :attr:`~toga.constants.Direction.HORIZONTAL`
        """
        super().__init__(id=id, style=style)

        # Create a platform specific implementation of a Divider
        self._impl = self.factory.Divider(interface=self)
        self.direction = direction

    @property
    def enabled(self) -> Literal[True]:
        """Is the widget currently enabled? i.e., can the user interact with the widget?

        Divider widgets cannot be disabled; this property will always return True; any
        attempt to modify it will be ignored.
        """
        return True

    @enabled.setter
    def enabled(self, value: object) -> None:
        pass

    def focus(self) -> None:
        """No-op; Divider cannot accept input focus."""
        pass

    @property
    def direction(self) -> Direction:
        """The direction in which the visual separator will be drawn."""
        return self._impl.get_direction()

    @direction.setter
    def direction(self, value: object) -> None:
        self._impl.set_direction(value)
        self.refresh()



from unittest.mock import Mock

from toga.sources import Source


def test_notify():
    source = Source()

    # No listeners
    assert source.listeners == []
    source.notify("message0")

    listener1 = Mock()

    source.add_listener(listener1)
    assert source.listeners == [listener1]

    # Add the same listener a second time
    source.add_listener(listener1)
    assert source.listeners == [listener1]

    # activate listener
    source.notify("message1")
    listener1.message1.assert_called_once_with()

    # activate listener with data
    source.notify("message2", arg1=11, arg2=22)
    listener1.message2.assert_called_once_with(arg1=11, arg2=22)

    # add more widgets to listeners
    listener2 = Mock()
    source.add_listener(listener2)
    assert source.listeners == [listener1, listener2]

    # activate listener
    source.notify("message3")
    listener1.message3.assert_called_once_with()
    listener2.message3.assert_called_once_with()

    # activate listener with data
    source.notify("message4", arg1=11, arg2=22)
    listener1.message4.assert_called_once_with(arg1=11, arg2=22)
    listener2.message4.assert_called_once_with(arg1=11, arg2=22)

    # remove listener2
    source.remove_listener(listener2)
    assert source.listeners == [listener1]

    # Activate listeners; listener2 not notified.
    source.notify("message5")
    listener1.message5.assert_called_once_with()
    listener2.message5.assert_not_called()


def test_missing_listener_method():
    """If a listener doesn't implement a notification method, the notification is
    ignored."""
    full_listener = Mock()
    partial_listener = object()
    source = Source()

    source.add_listener(full_listener)
    source.add_listener(partial_listener)
    assert source.listeners == [full_listener, partial_listener]

    # This shouldn't raise an error
    source.notify("message1")

    full_listener.message1.assert_called_once_with()



import os
import subprocess
import sys
from pathlib import Path

import toga


def run_app(args, cwd):
    """Run a Toga app as a subprocess with coverage enabled and the Toga Dummy
    backend."""
    # We need to do a full copy of the environment, then add our extra bits;
    # if we don't the Windows interpreter won't inherit SYSTEMROOT
    env = os.environ.copy()
    env.update(
        {
            "COVERAGE_PROCESS_START": str(
                Path(__file__).parent.parent / "pyproject.toml"
            ),
            "PYTHONPATH": str(Path(__file__).parent / "testbed/customize"),
            "TOGA_BACKEND": "toga_dummy",
        }
    )
    output = subprocess.check_output(
        [sys.executable] + args,
        cwd=cwd,
        env=env,
        text=True,
    )
    # When called as a subprocess, coverage drops its coverage report in CWD.
    # Move it to the project root for combination with the main test report.
    for file in cwd.glob(".coverage*"):
        os.rename(file, Path(__file__).parent.parent / file.name)
    return output


def assert_paths(output, app_path, app_name):
    """Assert the paths for the standalone app are consistent."""
    results = output.splitlines()
    assert f"app.paths.app={app_path.resolve()}" in results
    assert (
        f"app.paths.config={(Path.home() / 'config' / f'org.testbed.{app_name}').resolve()}"
        in results
    )
    assert (
        f"app.paths.data={(Path.home() / 'user_data' / f'org.testbed.{app_name}').resolve()}"
        in results
    )
    assert (
        f"app.paths.cache={(Path.home() / 'cache' / f'org.testbed.{app_name}').resolve()}"
        in results
    )
    assert (
        f"app.paths.logs={(Path.home() / 'logs' / f'org.testbed.{app_name}').resolve()}"
        in results
    )
    assert f"app.paths.toga={Path(toga.__file__).parent.resolve()}" in results


def test_as_interactive():
    """At an interactive prompt, the app path is the current working directory."""
    # Spawn the interactive-mode mocking entry point
    cwd = Path(__file__).parent / "testbed"
    output = run_app(["interactive.py"], cwd=cwd)
    assert_paths(output, app_path=cwd, app_name="interactive-app")


def test_simple_as_file_in_module():
    """When a simple app is started as `python app.py` inside a runnable module, the app
    path is the folder holding app.py."""
    # Spawn the simple testbed app using `app.py`
    cwd = Path(__file__).parent / "testbed/simple"
    output = run_app(["app.py"], cwd=cwd)
    assert_paths(output, app_path=Path(toga.__file__).parent, app_name="simple-app")


def test_simple_as_module():
    """When a simple apps is started as `python -m app` inside a runnable module, the
    app path is the folder holding app.py."""
    # Spawn the simple testbed app using `-m app`
    cwd = Path(__file__).parent / "testbed/simple"
    output = run_app(["-m", "app"], cwd=cwd)
    assert_paths(output, app_path=Path(toga.__file__).parent, app_name="simple-app")


def test_simple_as_deep_file():
    """When a simple app is started as `python simple/app.py`, the app path is the
    folder holding app.py."""
    # Spawn the simple testbed app using `simple/app.py`
    cwd = Path(__file__).parent / "testbed"
    output = run_app(["simple/app.py"], cwd=cwd)
    assert_paths(output, app_path=Path(toga.__file__).parent, app_name="simple-app")


def test_simple_as_deep_module():
    """When a simple app is started as `python -m simple`, the app path is the folder
    holding app.py."""
    # Spawn the simple testbed app using `-m simple`
    cwd = Path(__file__).parent / "testbed"
    output = run_app(["-m", "simple"], cwd=cwd)
    assert_paths(output, app_path=Path(toga.__file__).parent, app_name="simple-app")


def test_subclassed_as_file_in_module():
    """When a subclassed app is started as `python app.py` inside a runnable module, the
    app path is the folder holding app.py."""
    # Spawn the simple testbed app using `app.py`
    cwd = Path(__file__).parent / "testbed/subclassed"
    output = run_app(["app.py"], cwd=cwd)
    assert_paths(output, app_path=cwd, app_name="subclassed-app")


def test_subclassed_as_module():
    """When a subclassed app is started as `python -m app` inside a runnable module, the
    app path is the folder holding app.py."""
    # Spawn the subclassed testbed app using `-m app`
    cwd = Path(__file__).parent / "testbed/subclassed"
    output = run_app(["-m", "app"], cwd=cwd)
    assert_paths(output, app_path=cwd, app_name="subclassed-app")


def test_subclassed_as_deep_file():
    """When a subclassed app is started as `python simple/app.py`, the app path is the
    folder holding app.py."""
    # Spawn the subclassed testbed app using `subclassed/app.py`
    cwd = Path(__file__).parent / "testbed"
    output = run_app(["subclassed/app.py"], cwd=cwd)
    assert_paths(output, app_path=cwd / "subclassed", app_name="subclassed-app")


def test_subclassed_as_deep_module():
    """When a subclassed app is started as `python -m simple`, the app path is the
    folder holding app.py."""
    # Spawn the subclassed testbed app using `-m subclassed`
    cwd = Path(__file__).parent / "testbed"
    output = run_app(["-m", "subclassed"], cwd=cwd)
    assert_paths(output, app_path=cwd / "subclassed", app_name="subclassed-app")



from .accessors import to_accessor  # noqa: F401
from .base import Listener, Source  # noqa: F401
from .list_source import ListSource, Row  # noqa: F401
from .tree_source import Node, TreeSource  # noqa: F401
from .value_source import ValueSource  # noqa: F401

__all__ = [
    "ListSource",
    "Listener",
    "Node",
    "Row",
    "Source",
    "TreeSource",
    "ValueSource",
    "to_accessor",
]



import re
from unittest.mock import Mock

import pytest

from toga.sources import Node


def _create_node(source, parent, data, children=None):
    """A very simplified _create_node for mock purposes."""
    node = Node(**data)
    node._source = source
    node._parent = parent
    if children:
        node._children = [
            _create_node(source, parent=node, data=item[0], children=item[1])
            for item in children
        ]
    return node


@pytest.fixture
def source():
    source = Mock()
    source._accessors = ["val1", "val2"]
    source._create_node.side_effect = lambda *args, **kwargs: _create_node(
        source, *args, **kwargs
    )
    return source


@pytest.fixture
def leaf_node(source):
    node = Node(val1="value 1", val2=42)
    node._source = source
    return node


@pytest.fixture
def empty_node(source):
    node = Node()
    node._source = source
    node._children = []
    return node


@pytest.fixture
def child_a(source, leaf_node):
    child = Node(val1="value a", val2=111)
    child._source = source
    child._parent = leaf_node
    return child


@pytest.fixture
def child_b(source, leaf_node):
    child = Node(val1="value b", val2=222)
    child._source = source
    child._parent = leaf_node
    return child


@pytest.fixture
def node(leaf_node, child_a, child_b):
    leaf_node._children = [child_a, child_b]

    return leaf_node


def test_node_properties(node):
    """A node with children can be created and modified."""
    assert node.val1 == "value 1"
    assert node.val2 == 42
    assert node.can_have_children()
    assert len(node) == 2
    assert (
        re.match(r"<Node .* val1='value 1' val2=42; 2 children>", repr(node))
        is not None
    )


def test_empty_node_properties(empty_node):
    """An empty Node can be created."""
    assert empty_node.can_have_children()
    assert len(empty_node) == 0
    assert (
        re.match(r"<Node .* \(no attributes\); 0 children>", repr(empty_node))
        is not None
    )


def test_leaf_node_properties(leaf_node):
    """A Leaf Node can be created."""
    assert leaf_node.val1 == "value 1"
    assert leaf_node.val2 == 42
    assert not leaf_node.can_have_children()
    assert len(leaf_node) == 0
    assert (
        re.match(r"<Leaf Node .* val1='value 1' val2=42>", repr(leaf_node)) is not None
    )


def test_modify_attributes(source, node):
    """If node attributes are modified, a change notification is sent."""
    node.val1 = "new value"
    assert node.val1 == "new value"
    source.notify.assert_called_once_with("change", item=node)
    source.notify.reset_mock()

    # Deleting an attribute causes a change notification
    del node.val1
    assert not hasattr(node, "val1")
    source.notify.assert_called_once_with("change", item=node)
    source.notify.reset_mock()

    # Setting an attribute starting with with an underscore isn't a notifiable event
    node._secret = "secret value"
    assert node._secret == "secret value"
    source.notify.assert_not_called()

    # Deleting an attribute starting with with an underscore isn't a notifiable event
    del node._secret
    assert not hasattr(node, "_secret")
    source.notify.assert_not_called()

    # An attribute that wasn't in the original attribute set
    # still causes a change notification
    node.val3 = "other value"
    assert node.val3 == "other value"
    source.notify.assert_called_once_with("change", item=node)
    source.notify.reset_mock()

    # Deleting an attribute that wasn't in the original attribute set
    # still causes a change notification
    del node.val3
    assert not hasattr(node, "val3")
    source.notify.assert_called_once_with("change", item=node)
    source.notify.reset_mock()


def test_modify_attributes_no_source(node):
    """Node attributes can be modified on a node with no source."""
    node.source = None

    node.val1 = "new value"
    assert node.val1 == "new value"

    # Deleting an attribute causes a change notification
    del node.val1
    assert not hasattr(node, "val1")

    # Setting an attribute starting with with an underscore isn't a notifiable event
    node._secret = "secret value"
    assert node._secret == "secret value"

    # Deleting an attribute starting with with an underscore isn't a notifiable event
    del node._secret
    assert not hasattr(node, "_secret")

    # An attribute that wasn't in the original attribute set
    # still causes a change notification
    node.val3 = "other value"
    assert node.val3 == "other value"

    # Deleting an attribute that wasn't in the original attribute set
    # still causes a change notification
    del node.val3
    assert not hasattr(node, "val3")


def test_modify_children(source, node):
    """Node children can be retrieved and modified."""
    child = node[1]
    assert node[1].val1 == "value b"

    # Delete the child
    del node[1]

    # Removal notification was sent
    source.notify.assert_called_once_with("remove", parent=node, index=1, item=child)
    source.notify.reset_mock()

    # Child is no longer associated with the source
    assert child._parent is None
    assert child._source is None

    # Node child count has dropped
    assert len(node) == 1

    old_child_0 = node[0]

    # A child can be modified by index
    node[0] = {"val1": "new"}

    # Node 0 has changed instance
    assert old_child_0 is not node[0]

    # Old child 0 is no longer associated with this node
    assert old_child_0._source is None
    assert old_child_0._parent is None

    # A child node was created using the source's factory
    source._create_node.assert_called_with(parent=node, data={"val1": "new"})

    # Change notification was sent, the change is associated with the new item
    source.notify.assert_called_once_with("change", item=node[0])

    # Node child count hasn't changed
    assert len(node) == 1


def test_modify_leaf_children(leaf_node):
    """Attempts to modifying Leaf Node children raise an error."""
    with pytest.raises(
        ValueError,
        match=r"<Leaf Node .*> is a leaf node",
    ):
        leaf_node[1]

    with pytest.raises(
        ValueError,
        match=r"<Leaf Node .*> is a leaf node",
    ):
        del leaf_node[1]

    with pytest.raises(
        ValueError,
        match=r"<Leaf Node .*> is a leaf node",
    ):
        leaf_node[1] = {"val1": "new"}


def test_iterate(node):
    """Node can be iterated."""
    assert "|".join(child.val1 for child in node) == "value a|value b"


def test_iterate_leaf(leaf_node):
    """Node can be iterated."""
    assert "|".join(child.val1 for child in leaf_node) == ""


@pytest.mark.parametrize(
    "index, actual_index",
    [
        (1, 1),  # Positive, in range
        (10, 2),  # Positive, past positive limit
        (-1, 1),  # Negative, in range
        (-10, 0),  # Negative, past negative limit
    ],
)
def test_insert(source, node, index, actual_index):
    """A child can be inserted into a node."""
    new_child = node.insert(index, {"val1": "new"})

    # Node has one more child.
    assert len(node) == 3
    assert node[actual_index] == new_child

    # A child node was created using the source's factory
    source._create_node.assert_called_with(
        parent=node, data={"val1": "new"}, children=None
    )

    # insert notification was sent, the change is associated with the new item
    source.notify.assert_called_once_with(
        "insert",
        parent=node,
        index=actual_index,
        item=new_child,
    )


def test_insert_with_children(source, node):
    """A child with children can be inserted into a node."""
    new_child = node.insert(
        1,
        {"val1": "new"},
        children=[
            ({"val1": "new child 1"}, None),
            ({"val1": "new child 2"}, None),
        ],
    )

    # Node has one more child.
    assert len(node) == 3
    assert node[1] == new_child

    # A child node was created using the source's factory, and the child data was passed
    # to that call.
    source._create_node.assert_called_with(
        parent=node,
        data={"val1": "new"},
        children=[
            ({"val1": "new child 1"}, None),
            ({"val1": "new child 2"}, None),
        ],
    )

    # The children of the new child are as expected
    assert len(new_child) == 2
    assert new_child[0].val1 == "new child 1"
    assert not new_child[0].can_have_children()

    # insert notification was sent, the change is associated with the new item
    source.notify.assert_called_once_with(
        "insert",
        parent=node,
        index=1,
        item=new_child,
    )


def test_insert_leaf(leaf_node, source):
    """Inserting a child into a leaf makes the node not a leaf any more."""
    new_child = leaf_node.insert(0, {"val1": "new"})

    # Leaf node isn't a leaf any more
    assert leaf_node.can_have_children()
    assert len(leaf_node) == 1
    assert leaf_node[0] == new_child

    # A child node was created using the source's factory
    source._create_node.assert_called_with(
        parent=leaf_node, data={"val1": "new"}, children=None
    )

    # insert notification was sent, the change is associated with the new item
    source.notify.assert_called_once_with(
        "insert", parent=leaf_node, index=0, item=leaf_node[0]
    )


def test_append(source, node):
    """A child can be appended onto a node."""
    new_child = node.append({"val1": "new"})

    # Node has one more child.
    assert len(node) == 3
    assert node[2] == new_child

    # A child node was created using the source's factory
    source._create_node.assert_called_with(
        parent=node, data={"val1": "new"}, children=None
    )

    # insert notification was sent, the change is associated with the new item
    source.notify.assert_called_once_with(
        "insert",
        parent=node,
        index=2,
        item=new_child,
    )


def test_append_with_children(source, node):
    """A child with children can be appended onto a node."""
    new_child = node.append(
        {"val1": "new"},
        children=[
            ({"val1": "new child 1"}, None),
            ({"val1": "new child 2"}, None),
        ],
    )

    # Node has one more child.
    assert len(node) == 3
    assert node[2] == new_child

    # A child node was created using the source's factory, and the child data was passed
    # to that call. Since our source is a mock, we won't get actual children.
    source._create_node.assert_called_with(
        parent=node,
        data={"val1": "new"},
        children=[
            ({"val1": "new child 1"}, None),
            ({"val1": "new child 2"}, None),
        ],
    )

    # insert notification was sent, the change is associated with the new item
    source.notify.assert_called_once_with(
        "insert",
        parent=node,
        index=2,
        item=new_child,
    )


def test_append_leaf(leaf_node, source):
    """Appending to a leaf makes the node not a leaf any more."""
    new_child = leaf_node.append({"val1": "new"})

    # Leaf node isn't a leaf any more
    assert leaf_node.can_have_children()
    assert len(leaf_node) == 1
    assert leaf_node[0] == new_child

    # A child node was created using the source's factory
    source._create_node.assert_called_with(
        parent=leaf_node, data={"val1": "new"}, children=None
    )

    # insert notification was sent, the change is associated with the new item
    source.notify.assert_called_once_with(
        "insert", parent=leaf_node, index=0, item=leaf_node[0]
    )


def test_index(node, child_b):
    """A child can be found it it's parent."""
    assert node.index(child_b) == 1


def test_index_not_child(node):
    """If a node isn't a child of this node, it can't be found by index."""
    other = Node(val1="other")

    with pytest.raises(
        ValueError,
        match=r"<Leaf Node .* val1='other'> is not in list",
    ):
        node.index(other)


def test_index_leaf(leaf_node, child_b):
    """A child cannot be found in a leaf node."""
    with pytest.raises(
        ValueError,
        match=r"<Leaf Node .* val1='value 1' val2=42> is a leaf node",
    ):
        leaf_node.index(child_b)


def test_remove(source, node, child_b):
    """A node can be removed from it's parent."""
    # Child is initially associated with the node
    assert child_b._parent == node
    assert child_b._source == node._source

    # Remove the child
    node.remove(child_b)

    # Child isn't associated with the node any more
    assert child_b._parent is None
    assert child_b._source is None

    # The node has less children
    assert len(node) == 1

    # The source was notified
    source.notify.assert_called_once_with("remove", parent=node, index=1, item=child_b)


def test_remove_leaf(leaf_node, child_b):
    """A child cannot be removed from a leaf node."""
    with pytest.raises(
        ValueError,
        match=r"<Leaf Node .* val1='value 1' val2=42> is a leaf node",
    ):
        leaf_node.index(child_b)


def test_find(node, child_b):
    """A node can be found by value in it's parent's list of children."""
    # Append some additional children
    child_c = node.append({"val1": "value a", "val2": 333})
    child_d = node.append({"val1": "value b", "val2": 444})

    # Find the child by a partial match of values.
    assert node.find({"val1": "value b"}) == child_b

    # Find the child by a partial match of values, starting at the first match
    assert node.find({"val1": "value b"}, start=child_b) == child_d

    # Find the child by a full match of values, starting at the first match
    assert node.find({"val1": "value a", "val2": 333}) == child_c


def test_found_leaf(leaf_node):
    """A child cannot be found from a leaf node."""
    with pytest.raises(
        ValueError,
        match=r"<Leaf Node .* val1='value 1' val2=42> is a leaf node",
    ):
        leaf_node.find({"val1": "value 1"})



import pytest

import toga
from toga.app import WindowSet


@pytest.fixture
def window1(app):
    return toga.Window(title="Window 1", id="a-win")


@pytest.fixture
def window2(app):
    return toga.Window(title="Window 2", id="z-win")


def test_create(app):
    """An empty windowset can be created."""
    windowset = WindowSet(app)

    assert windowset.app == app
    assert len(windowset) == 0


def test_add_discard(app, window1, window2):
    """An item can be added to a windowset."""
    # The windowset has 3 windows - the main window, plus 2 extras
    assert len(app.windows) == 3

    # Check the iterator and sorting works
    assert sorted(iter(app.windows)) == [window1, app.main_window, window2]

    with pytest.raises(
        TypeError,
        match=r"Can only add objects of type toga.Window",
    ):
        app.windows.add(object())

    # Explicitly re-add a window that is already in the windowset
    app.windows.add(window2)
    assert len(app.windows) == 3
    assert window2 in app.windows
    assert window2.app == app

    # Explicitly discard a window that is in the windowset
    app.windows.discard(window2)
    assert window2 not in app.windows

    # Duplicate discard - it's no longer a member
    with pytest.raises(
        ValueError,
        match=r"<toga.window.Window object at .*> is not part of this app",
    ):
        app.windows.discard(window2)

    with pytest.raises(
        TypeError,
        match=r"Can only discard objects of type toga.Window",
    ):
        app.windows.discard(object())


def test_iadd_isub(app, window1, window2):
    """The deprecated += and -= operators are no-ops."""
    # The windowset has 3 windows - the main window, plus 2 extras
    assert window2 in app.windows
    assert len(app.windows) == 3

    with pytest.warns(
        DeprecationWarning,
        match=r"Windows are automatically associated with the app; \+= is not required",
    ):
        app.windows += window2

    assert window2 in app.windows
    assert len(app.windows) == 3

    with pytest.warns(
        DeprecationWarning,
        match=r"Windows are automatically removed from the app; -= is not required",
    ):
        app.windows -= window2

    # -= is a no-op.
    assert window2 in app.windows
    assert len(app.windows) == 3

    with pytest.raises(AttributeError, match=r"can't set attribute 'windows'"):
        app.windows = None



from __future__ import annotations

import re
from collections.abc import Iterable, Mapping

NON_ACCESSOR_CHARS = re.compile(r"[^\w ]")
WHITESPACE = re.compile(r"\s+")


def to_accessor(heading: str) -> str:
    """Convert a human-readable heading into a data attribute accessor.

    This is done by:

    1. Converting the heading to lower case;
    2. Removing any character that can't be used in a Python identifier
    3. Replacing all whitespace with "_"
    4. Prepending ``_`` if the first character is a digit.

    Examples:

    * 'Heading 1' -> 'heading_1'
    * 'Heading - Title' -> 'heading_title'
    * 'Heading!' -> 'heading'
    * '1 Heading' -> '_1_heading'
    * '你好' -> '你好'

    :param heading: The column heading.
    :returns: The accessor derived from the heading.
    :raises ValueError: If the heading cannot be converted into an accessor.
    """
    value = WHITESPACE.sub(
        " ",
        NON_ACCESSOR_CHARS.sub("", heading.lower()),
    ).replace(" ", "_")

    try:
        if value[0].isdigit():
            value = f"_{value}"
    except IndexError:
        raise ValueError(
            f"Unable to automatically generate accessor from heading {heading!r}."
        )

    return value


def build_accessors(
    headings: Iterable[str],
    accessors: Iterable[str | None] | Mapping[str, str] | None,
) -> list[str]:
    """Convert a list of headings (with accessor overrides) to a finalised list of
    accessors.

    :param headings: The list of headings.
    :param accessors: The list of accessor overrides. Can be specified as:

        * A list the same length as headings. Each entry in the list is a string that
          is the override name for the accessor, or :any:`None` if the default accessor
          for the heading at that index should be used.
        * A dictionary mapping heading names to accessor names. If a heading name isn't
          present in the dictionary, the default accessor will be used.

    :returns: The final list of accessors.
    """
    if accessors is not None:
        if isinstance(accessors, Mapping):
            result = [
                accessors[h] if h in accessors else to_accessor(h) for h in headings
            ]
        else:
            # TODO: use zip(..., strict=True) instead once Python 3.9 support is dropped
            if len(headings := list(headings)) != len(accessors := list(accessors)):
                raise ValueError("Number of accessors must match number of headings")

            result = [
                a if a is not None else to_accessor(h)
                for h, a in zip(headings, accessors)
            ]
    else:
        result = [to_accessor(h) for h in headings]

    return result



import pytest

import toga


@pytest.fixture
def parent_group_1():
    return toga.Group("P1", order=1)


@pytest.fixture
def child_group_1(parent_group_1):
    return toga.Group("C1", order=2, parent=parent_group_1)


@pytest.fixture
def child_group_2(parent_group_1):
    return toga.Group("C2", order=4, parent=parent_group_1)


@pytest.fixture
def parent_group_2():
    return toga.Group("P2", order=2)


@pytest.fixture
def child_group_3(parent_group_2):
    return toga.Group("C3", section=2, order=2, parent=parent_group_2)


@pytest.fixture
def child_group_4(parent_group_2):
    return toga.Group("C4", section=2, order=1, parent=parent_group_2)


@pytest.fixture
def child_group_5(parent_group_2):
    return toga.Group("C5", section=1, order=1, parent=parent_group_2)



import pytest

import toga
from toga.app import WidgetRegistry


@pytest.fixture
def widget_registry():
    return WidgetRegistry()


# Create the simplest possible widget with a concrete implementation
class ExampleWidget(toga.Widget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._impl = self.factory.Widget(self)

    def __repr__(self):
        return f"Widget(id={self.id!r})"


def test_empty_registry(widget_registry):
    assert len(widget_registry) == 0
    assert list(widget_registry) == []
    assert str(widget_registry) == "{}"


def test_add_widget(widget_registry):
    """Widgets can be added to the registry."""
    # Add a widget to the registry
    widget1 = ExampleWidget(id="widget-1")
    widget_registry._add(widget1)

    assert len(widget_registry) == 1
    assert list(widget_registry) == [widget1]
    assert str(widget_registry) == "{'widget-1': Widget(id='widget-1')}"
    assert widget_registry["widget-1"] == widget1

    # Add a second widget
    widget2 = ExampleWidget(id="widget-2")
    widget_registry._add(widget2)

    assert len(widget_registry) == 2
    assert widget_registry["widget-1"] == widget1
    assert widget_registry["widget-2"] == widget2


def test_update_widgets(widget_registry):
    """The registry can be bulk updated."""
    # Add a widget to the registry
    widget1 = ExampleWidget(id="widget-1")
    widget_registry._add(widget1)

    widget2 = ExampleWidget(id="widget-2")
    widget3 = ExampleWidget(id="widget-3")
    widget4 = ExampleWidget(id="widget-4")
    widget_registry._update({widget2, widget3, widget4})

    assert len(widget_registry) == 4
    assert widget_registry["widget-1"] == widget1
    assert widget_registry["widget-2"] == widget2
    assert widget_registry["widget-3"] == widget3
    assert widget_registry["widget-4"] == widget4


def test_remove_widget(widget_registry):
    """A widget can be removed from the repository."""
    "Widgets can be added to the registry"
    # Add a widget to the registry
    widget1 = ExampleWidget(id="widget-1")
    widget2 = ExampleWidget(id="widget-2")
    widget_registry._update({widget1, widget2})

    assert len(widget_registry) == 2

    widget_registry._remove("widget-2")

    assert widget_registry["widget-1"] == widget1
    assert "widget-2" not in widget_registry


def test_add_same_widget_twice(widget_registry):
    """A widget cannot be added to the same registry twice."""
    # Add a widget to the registry
    widget1 = ExampleWidget(id="widget-1")
    widget_registry._add(widget1)

    assert len(widget_registry) == 1

    # Add the widget again; this raises an error
    with pytest.raises(
        KeyError,
        match=r"There is already a widget with the id 'widget-1'",
    ):
        widget_registry._add(widget1)

    # Widget is still there
    assert len(widget_registry) == 1
    assert widget_registry["widget-1"] == widget1


def test_add_duplicate_id(widget_registry):
    """A widget cannot be added to the same registry twice."""
    # Add a widget to the registry
    widget1 = ExampleWidget(id="widget-1")
    widget_registry._add(widget1)

    assert len(widget_registry) == 1

    new_widget = ExampleWidget(id="widget-1")

    # Add the widget again; this raises an error
    with pytest.raises(
        KeyError,
        match=r"There is already a widget with the id 'widget-1'",
    ):
        widget_registry._add(new_widget)

    # Widget is still there
    assert len(widget_registry) == 1
    assert widget_registry["widget-1"] == widget1


def test_setitem(widget_registry):
    """Widgets cannot be directly assigned to the registry."""
    widget1 = ExampleWidget(id="widget-1")

    with pytest.raises(
        TypeError,
        match=r"'WidgetRegistry' object does not support item assignment",
    ):
        widget_registry["new is"] = widget1



from toga.style.pack import BOTTOM, CENTER, COLUMN, ROW, TOP, Pack

from ..utils import ExampleNode, ExampleViewport, assert_layout


def test_top():
    root = ExampleNode(
        "root",
        style=Pack(direction=ROW, alignment=TOP, width=300),
        children=[
            ExampleNode("space_filler", style=Pack(width=100, height=540)),
            ExampleNode(
                "container",
                style=Pack(direction=COLUMN, padding_top=110, padding_bottom=120),
                children=[
                    ExampleNode(
                        "widget_a",
                        style=Pack(height=30, padding_top=32, padding_bottom=34),
                    ),
                    ExampleNode(
                        "widget_b",
                        style=Pack(height=36, padding_top=38, padding_bottom=40),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (300, 540),
        (300, 540),
        {
            "origin": (0, 0),
            "content": (300, 540),
            "children": [
                {"origin": (0, 0), "content": (100, 540)},
                {
                    "origin": (100, 110),
                    "content": (200, 210),
                    "children": [
                        {"origin": (100, 142), "content": (200, 30)},
                        {"origin": (100, 244), "content": (200, 36)},
                    ],
                },
            ],
        },
    )


def test_center():
    root = ExampleNode(
        "root",
        style=Pack(direction=ROW, alignment=CENTER, width=300),
        children=[
            ExampleNode("space_filler", style=Pack(width=100, height=540)),
            ExampleNode(
                "container",
                style=Pack(direction=COLUMN, padding_top=110, padding_bottom=120),
                children=[
                    ExampleNode(
                        "widget_a",
                        style=Pack(height=30, padding_top=32, padding_bottom=34),
                    ),
                    ExampleNode(
                        "widget_b",
                        style=Pack(height=36, padding_top=38, padding_bottom=40),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (300, 540),
        (300, 540),
        {
            "origin": (0, 0),
            "content": (300, 540),
            "children": [
                {"origin": (0, 0), "content": (100, 540)},
                {
                    "origin": (100, 160),
                    "content": (200, 210),
                    "children": [
                        {"origin": (100, 192), "content": (200, 30)},
                        {"origin": (100, 294), "content": (200, 36)},
                    ],
                },
            ],
        },
    )


def test_bottom():
    root = ExampleNode(
        "root",
        style=Pack(direction=ROW, alignment=BOTTOM, width=300),
        children=[
            ExampleNode("space_filler", style=Pack(width=100, height=540)),
            ExampleNode(
                "container",
                style=Pack(direction=COLUMN, padding_top=110, padding_bottom=120),
                children=[
                    ExampleNode(
                        "widget_a",
                        style=Pack(height=30, padding_top=32, padding_bottom=34),
                    ),
                    ExampleNode(
                        "widget_b",
                        style=Pack(height=36, padding_top=38, padding_bottom=40),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (300, 540),
        (300, 540),
        {
            "origin": (0, 0),
            "content": (300, 540),
            "children": [
                {"origin": (0, 0), "content": (100, 540)},
                {
                    "origin": (100, 210),
                    "content": (200, 210),
                    "children": [
                        {"origin": (100, 242), "content": (200, 30)},
                        {"origin": (100, 344), "content": (200, 36)},
                    ],
                },
            ],
        },
    )


def test_no_padding():
    root = ExampleNode(
        "root",
        style=Pack(direction=ROW, alignment=CENTER, width=300),
        children=[
            ExampleNode("space_filler", style=Pack(width=100, height=540)),
            ExampleNode("widget", style=Pack(height=440)),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (300, 540),
        (300, 540),
        {
            "origin": (0, 0),
            "content": (300, 540),
            "children": [
                {"origin": (0, 0), "content": (100, 540)},
                {"origin": (100, 50), "content": (200, 440)},
            ],
        },
    )


def test_row_box():
    root = ExampleNode(
        "root",
        style=Pack(direction=ROW, alignment=CENTER),
        children=[
            ExampleNode("space_filler", style=Pack(width=100, height=430)),
            ExampleNode(
                "container",
                style=Pack(direction=ROW, padding_top=110, padding_bottom=120),
                children=[
                    ExampleNode(
                        "widget_a",
                        style=Pack(
                            width=30, height=100, padding_left=32, padding_right=34
                        ),
                    ),
                    ExampleNode(
                        "widget_b",
                        style=Pack(
                            width=36, height=100, padding_left=38, padding_right=40
                        ),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (310, 430),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {"origin": (0, 25), "content": (100, 430)},
                {
                    "origin": (100, 110),
                    "content": (210, 250),
                    "children": [
                        {"origin": (132, 110), "content": (30, 100)},
                        {"origin": (234, 110), "content": (36, 100)},
                    ],
                },
            ],
        },
    )



from __future__ import annotations

from collections.abc import Iterable, Mapping, Sequence
from typing import Generic, TypeVar

from .base import Source

T = TypeVar("T")


def _find_item(
    candidates: Sequence[T],
    data: object,
    accessors: Sequence[str],
    start: T | None,
    error: str,
) -> T:
    """Find-by-value implementation helper; find an item matching ``data`` in
    ``candidates``, starting with item ``start``."""
    if start is not None:
        start_index = candidates.index(start) + 1
    else:
        start_index = 0

    for item in candidates[start_index:]:
        try:
            if isinstance(data, Mapping):
                found = all(
                    getattr(item, attr) == value for attr, value in data.items()
                )
            elif hasattr(data, "__iter__") and not isinstance(data, str):
                found = all(
                    getattr(item, attr) == value for value, attr in zip(data, accessors)
                )
            else:
                found = getattr(item, accessors[0]) == data

            if found:
                return item
        except AttributeError:
            # Attribute didn't exist, so it's not a match
            pass

    raise ValueError(error)


class Row(Generic[T]):
    def __init__(self, **data: T):
        """Create a new Row object.

        The keyword arguments specified in the constructor will be converted into
        attributes on the new Row object.

        When any public attributes of the Row are modified (i.e., any attribute whose
        name doesn't start with ``_``), the source to which the row belongs will be
        notified.
        """
        self._source: Source | None = None
        for name, value in data.items():
            setattr(self, name, value)

    def __repr__(self) -> str:
        descriptor = " ".join(
            f"{attr}={getattr(self, attr)!r}"
            for attr in sorted(self.__dict__)
            if not attr.startswith("_")
        )
        return f"<Row {id(self):x} {descriptor if descriptor else '(no attributes)'}>"

    ######################################################################
    # Utility wrappers
    ######################################################################

    def __setattr__(self, attr: str, value: T) -> None:
        """Set an attribute on the Row object, notifying the source of the change.

        :param attr: The attribute to change.
        :param value: The new attribute value.
        """
        super().__setattr__(attr, value)
        if not attr.startswith("_"):
            if self._source is not None:
                self._source.notify("change", item=self)

    def __getattr__(self, attr: str) -> T:
        return super().__getattr__(attr)

    def __delattr__(self, attr: str) -> None:
        """Remove an attribute from the Row object, notifying the source of the change.

        :param attr: The attribute to change.
        """
        super().__delattr__(attr)
        if not attr.startswith("_"):
            if self._source is not None:
                self._source.notify("change", item=self)


class ListSource(Source):
    _data: list[Row]

    def __init__(self, accessors: Iterable[str], data: Iterable | None = None):
        """A data source to store an ordered list of multiple data values.

        :param accessors: A list of attribute names for accessing the value
            in each column of the row.
        :param data: The initial list of items in the source. Items are converted as
            shown :ref:`above <listsource-item>`.
        """
        super().__init__()
        if isinstance(accessors, str) or not hasattr(accessors, "__iter__"):
            raise ValueError("accessors should be a list of attribute names")

        # Copy the list of accessors
        self._accessors = [a for a in accessors]
        if len(self._accessors) == 0:
            raise ValueError("ListSource must be provided a list of accessors")

        # Convert the data into row objects
        if data is not None:
            self._data = [self._create_row(value) for value in data]
        else:
            self._data = []

    ######################################################################
    # Methods required by the ListSource interface
    ######################################################################

    def __len__(self) -> int:
        """Returns the number of items in the list."""
        return len(self._data)

    def __getitem__(self, index: int) -> Row:
        """Returns the item at position ``index`` of the list."""
        return self._data[index]

    def __delitem__(self, index: int) -> None:
        """Deletes the item at position ``index`` of the list."""
        row = self._data[index]
        del self._data[index]
        self.notify("remove", index=index, item=row)

    ######################################################################
    # Factory methods for new rows
    ######################################################################

    # This behavior is documented in list_source.rst.
    def _create_row(self, data: object) -> Row:
        if isinstance(data, Mapping):
            row = Row(**data)
        elif hasattr(data, "__iter__") and not isinstance(data, str):
            row = Row(**dict(zip(self._accessors, data)))
        else:
            row = Row(**{self._accessors[0]: data})
        row._source = self
        return row

    ######################################################################
    # Utility methods to make ListSources more list-like
    ######################################################################

    def __setitem__(self, index: int, value: object) -> None:
        """Set the value of a specific item in the data source.

        :param index: The item to change
        :param value: The data for the updated item. This data will be converted
            into a Row object.
        """
        row = self._create_row(value)
        self._data[index] = row
        self.notify("insert", index=index, item=row)

    def clear(self) -> None:
        """Clear all data from the data source."""
        self._data = []
        self.notify("clear")

    def insert(self, index: int, data: object) -> Row:
        """Insert a row into the data source at a specific index.

        :param index: The index at which to insert the item.
        :param data: The data to insert into the ListSource. This data will be converted
            into a Row object.
        :returns: The newly constructed Row object.
        """
        row = self._create_row(data)
        self._data.insert(index, row)
        self.notify("insert", index=index, item=row)
        return row

    def append(self, data: object) -> Row:
        """Insert a row at the end of the data source.

        :param data: The data to append to the ListSource. This data will be converted
            into a Row object.
        :returns: The newly constructed Row object.
        """
        return self.insert(len(self), data)

    def remove(self, row: Row) -> None:
        """Remove a row from the data source.

        :param row: The row to remove from the data source.
        """
        del self[self._data.index(row)]

    def index(self, row: Row) -> int:
        """The index of a specific row in the data source.

        This search uses Row instances, and searches for an *instance* match.
        If two Row instances have the same values, only the Row that is the
        same Python instance will match. To search for values based on equality,
        use :meth:`~toga.sources.ListSource.find`.

        :param row: The row to find in the data source.
        :returns: The index of the row in the data source.
        :raises ValueError: If the row cannot be found in the data source.
        """
        return self._data.index(row)

    def find(self, data: object, start: Row | None = None) -> Row:
        """Find the first item in the data that matches all the provided
        attributes.

        This is a value based search, rather than an instance search. If two Row
        instances have the same values, the first instance that matches will be
        returned. To search for a second instance, provide the first found instance
        as the ``start`` argument. To search for a specific Row instance, use the
        :meth:`~toga.sources.ListSource.index`.

        :param data: The data to search for. Only the values specified in data will be
            used as matching criteria; if the row contains additional data attributes,
            they won't be considered as part of the match.
        :param start: The instance from which to start the search. Defaults to ``None``,
            indicating that the first match should be returned.
        :return: The matching Row object
        :raises ValueError: If no match is found.
        """
        return _find_item(
            candidates=self._data,
            data=data,
            accessors=self._accessors,
            start=start,
            error=f"No row matching {data!r} in data",
        )



from __future__ import annotations

import asyncio
import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from toga.app import App
    from toga.window import Window


class Document(ABC):
    def __init__(
        self,
        path: str | Path,
        document_type: str,
        app: App,
    ):
        """Create a new Document.

        :param path: The path where the document is stored.
        :param document_type: A human-readable description of the document type.
        :param app: The application the document is associated with.
        """
        self._path = Path(path)
        self._document_type = document_type
        self._app = app
        self._main_window: Window | None = None

        # Create the visual representation of the document
        self.create()

        # Create a platform specific implementation of the Document
        self._impl = app.factory.Document(interface=self)

    def can_close(self) -> bool:
        """Is the main document window allowed to close?

        The default implementation always returns ``True``; subclasses can override this
        to prevent a window closing with unsaved changes, etc.

        This default implementation is a function; however, subclasses can define it
        as an asynchronous co-routine if necessary to allow for dialog confirmations.
        """
        return True

    async def handle_close(self, window: Window, **kwargs: object) -> bool:
        """An ``on-close`` handler for the main window of this document that implements
        platform-specific document close behavior.

        It interrogates the :meth:`~toga.Document.can_close()` method to determine if
        the document is allowed to close.
        """
        if asyncio.iscoroutinefunction(self.can_close):
            can_close = await self.can_close()
        else:
            can_close = self.can_close()

        if can_close:
            if self._impl.SINGLE_DOCUMENT_APP:
                self.app.exit()
                return False
            else:
                return True
        else:
            return False

    @property
    def path(self) -> Path:
        """The path where the document is stored (read-only)."""
        return self._path

    @property
    def filename(self) -> Path:
        """**DEPRECATED** - Use :attr:`path`."""
        warnings.warn(
            "Document.filename has been renamed Document.path.",
            DeprecationWarning,
        )
        return self._path

    @property
    def document_type(self) -> str:
        """A human-readable description of the document type (read-only)."""
        return self._document_type

    @property
    def app(self) -> App:
        """The app that this document is associated with (read-only)."""
        return self._app

    @property
    def main_window(self) -> Window | None:
        """The main window for the document."""
        return self._main_window

    @main_window.setter
    def main_window(self, window: Window) -> None:
        self._main_window = window

    def show(self) -> None:
        """Show the :any:`main_window` for this document."""
        self.main_window.show()

    @abstractmethod
    def create(self) -> None:
        """Create the window (or windows) for the document.

        This method must, at a minimum, assign the :any:`main_window` property. It
        may also create additional windows or UI elements if desired.
        """

    @abstractmethod
    def read(self) -> None:
        """Load a representation of the document into memory and populate the document
        window."""



import pytest

import toga


# Create the simplest possible widget with a concrete implementation
class ExampleWidget(toga.Widget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._impl = self.factory.Widget(self)

    def __repr__(self):
        return f"Widget(id={self.id!r})"


# Create a box subclass with a reproducible repr
class ExampleBox(toga.Box):
    def __repr__(self):
        return f"Box(id={self.id!r})"


def make_window(win_id):
    win = toga.Window(id=win_id)
    win.content = ExampleBox(
        id=f"{win_id}.0",
        children=[
            ExampleWidget(id=f"{win_id}.1"),
            ExampleWidget(id=f"{win_id}.2"),
        ],
    )
    return win


def test_membership(app):
    """The registry proxy returns the total widget count."""
    # Create 2 windows with widgets
    win_1 = make_window("1")
    win_2 = make_window("2")

    # The widget registry now has 6 items in it
    assert len(app.widgets) == 6
    assert "1.0" in app.widgets
    assert "1.1" in app.widgets
    assert "1.2" in app.widgets
    assert "1.X" not in app.widgets
    assert "2.0" in app.widgets
    assert "2.1" in app.widgets
    assert "2.2" in app.widgets
    assert "2.X" not in app.widgets

    assert app.widgets["1.1"] == win_1.content.children[0]
    assert app.widgets["2.1"] == win_2.content.children[0]

    assert sorted(app.widgets.keys()) == ["1.0", "1.1", "1.2", "2.0", "2.1", "2.2"]
    assert sorted(app.widgets.values()) == [
        win_1.content,
        win_1.content.children[0],
        win_1.content.children[1],
        win_2.content,
        win_2.content.children[0],
        win_2.content.children[1],
    ]
    assert sorted(iter(app.widgets)) == [
        win_1.content,
        win_1.content.children[0],
        win_1.content.children[1],
        win_2.content,
        win_2.content.children[0],
        win_2.content.children[1],
    ]
    assert sorted(app.widgets.items()) == [
        ("1.0", win_1.content),
        ("1.1", win_1.content.children[0]),
        ("1.2", win_1.content.children[1]),
        ("2.0", win_2.content),
        ("2.1", win_2.content.children[0]),
        ("2.2", win_2.content.children[1]),
    ]
    assert repr(app.widgets) == (
        "{'1.0': Box(id='1.0'), '1.1': Widget(id='1.1'), '1.2': Widget(id='1.2'), "
        "'2.0': Box(id='2.0'), '2.1': Widget(id='2.1'), '2.2': Widget(id='2.2')}"
    )

    # Window 1's registry only has the subset of win1 widgets
    assert len(win_1.widgets) == 3
    assert "1.0" in win_1.widgets
    assert "1.1" in win_1.widgets
    assert "1.2" in win_1.widgets
    assert "1.X" not in win_1.widgets
    assert "2.0" not in win_1.widgets
    assert "2.1" not in win_1.widgets
    assert "2.2" not in win_1.widgets
    assert "2.X" not in win_1.widgets

    assert win_1.widgets["1.1"] == win_1.content.children[0]

    with pytest.raises(KeyError, match=r"2\.1"):
        win_1.widgets["2.1"]

    assert sorted(win_1.widgets.keys()) == ["1.0", "1.1", "1.2"]
    assert sorted(win_1.widgets.values()) == [
        win_1.content,
        win_1.content.children[0],
        win_1.content.children[1],
    ]
    assert sorted(iter(win_1.widgets)) == [
        win_1.content,
        win_1.content.children[0],
        win_1.content.children[1],
    ]
    assert sorted(win_1.widgets.items()) == [
        ("1.0", win_1.content),
        ("1.1", win_1.content.children[0]),
        ("1.2", win_1.content.children[1]),
    ]
    assert (
        repr(win_1.widgets)
        == "{'1.0': Box(id='1.0'), '1.1': Widget(id='1.1'), '1.2': Widget(id='1.2')}"
    )

    # Win 2 is the same as win 1; do a quick check
    assert len(win_2.widgets) == 3
    assert (
        repr(win_2.widgets)
        == "{'2.0': Box(id='2.0'), '2.1': Widget(id='2.1'), '2.2': Widget(id='2.2')}"
    )

    # Remove a widget
    widget = win_1.widgets["1.1"]
    win_1.content.remove(widget)

    # It is no longer in the app registry.
    assert len(app.widgets) == 5
    assert len(win_1.widgets) == 2
    assert "1.1" not in app.widgets
    assert "1.1" not in win_1.widgets

    with pytest.raises(KeyError, match=r"1\.1"):
        app.widgets["1.1"]

    with pytest.raises(KeyError, match=r"1\.1"):
        win_1.widgets["1.1"]


def test_reuse_id(app):
    """A widget ID can be reused once a widget has been removed (#2190)"""
    # Create 2 windows with widgets
    win_1 = make_window("1")
    win_2 = make_window("2")

    # the `magic` ID doesn't exist initially
    assert "magic" not in app.widgets
    assert "magic" not in win_1.widgets
    assert "magic" not in win_2.widgets

    # Create a widget with the magic ID. The widget still isn't in the registry, because
    # it's not part of a window
    first = ExampleWidget(id="magic")

    assert "magic" not in app.widgets
    assert "magic" not in win_1.widgets
    assert "magic" not in win_2.widgets

    # Add the widget to the window. This adds it to the registry
    win_1.content.add(first)

    assert "magic" in app.widgets
    assert "magic" in win_1.widgets
    assert "magic" not in win_2.widgets
    assert app.widgets["magic"] == first

    # Create a second widget with the same ID.
    # The widget exists, but the registry is storing `first`.
    second = ExampleWidget(id="magic")

    assert "magic" in app.widgets
    assert "magic" in win_1.widgets
    assert "magic" not in win_2.widgets
    assert app.widgets["magic"] == first

    # The second widget can't be added as content to either window, because of the ID clash.
    with pytest.raises(
        KeyError,
        match=r"There is already a widget with the id 'magic'",
    ):
        win_1.content.add(second)

    assert "magic" in app.widgets
    assert "magic" in win_1.widgets
    assert "magic" not in win_2.widgets
    assert app.widgets["magic"] == first

    # The widget can't be added to a different window, because there's still an app-level conflict.
    with pytest.raises(
        KeyError,
        match=r"There is already a widget with the id 'magic'",
    ):
        win_2.content.add(second)

    assert "magic" in app.widgets
    assert "magic" in win_1.widgets
    assert "magic" not in win_2.widgets
    assert app.widgets["magic"] == first

    # Remove the first widget
    win_1.content.remove(first)

    assert "magic" not in app.widgets
    assert "magic" not in win_1.widgets
    assert "magic" not in win_2.widgets
    with pytest.raises(KeyError, match=r"magic"):
        app.widgets["magic"]

    # The second widget can now be added to as content.
    win_1.content.add(second)

    assert "magic" in app.widgets
    assert "magic" in win_1.widgets
    assert "magic" not in win_2.widgets
    assert app.widgets["magic"] == second



from unittest.mock import Mock

import pytest

import toga
from toga_dummy import factory as dummy_factory
from toga_dummy.utils import (
    EventLog,
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
)


@pytest.fixture
def window():
    return toga.Window()


@pytest.fixture
def content1():
    return toga.Box()


@pytest.fixture
def content2():
    return toga.Box()


@pytest.fixture
def content3():
    return toga.Box()


@pytest.fixture
def content4():
    return toga.Box()


@pytest.fixture
def on_select_handler():
    return Mock()


@pytest.fixture
def tab_icon(app):
    return toga.Icon("tab-icon")


@pytest.fixture
def optioncontainer(
    content1,
    content2,
    content3,
    content4,
    on_select_handler,
    tab_icon,
):
    return toga.OptionContainer(
        content=[
            ("Item 1", content1),
            ("Item 2", content2, "other-icon"),
            ("Item 3", content3, tab_icon, False),
            toga.OptionItem("Item 4", content4),
        ],
        on_select=on_select_handler,
    )


def test_widget_create():
    """An option container can be created with no arguments."""
    optioncontainer = toga.OptionContainer()
    assert_action_performed(optioncontainer, "create OptionContainer")

    assert len(optioncontainer.content) == 0
    assert optioncontainer.current_tab is None
    assert optioncontainer.on_select._raw is None


def test_widget_create_with_args(
    optioncontainer,
    content1,
    content2,
    content3,
    content4,
    on_select_handler,
    tab_icon,
):
    """An option container can be created with arguments."""
    assert optioncontainer._impl.interface == optioncontainer
    assert_action_performed(optioncontainer, "create OptionContainer")

    assert len(optioncontainer.content) == 4
    assert optioncontainer.current_tab.text == "Item 1"
    assert optioncontainer.current_tab.icon is None
    assert optioncontainer.current_tab.enabled
    assert optioncontainer.current_tab.content == content1
    assert optioncontainer.on_select._raw == on_select_handler

    assert optioncontainer.content[1].text == "Item 2"
    assert optioncontainer.content[1].icon.path.name == "other-icon"
    assert optioncontainer.content[1].content == content2
    assert optioncontainer.content[1].enabled

    assert optioncontainer.content[2].text == "Item 3"
    assert optioncontainer.content[2].icon == tab_icon
    assert optioncontainer.content[2].content == content3
    assert not optioncontainer.content[2].enabled

    assert optioncontainer.content[3].text == "Item 4"
    assert optioncontainer.content[3].icon is None
    assert optioncontainer.content[3].content == content4
    assert optioncontainer.content[3].enabled


@pytest.mark.parametrize(
    "value",
    [
        ("label",),
        ("label", toga.Box(), None, True, "extra"),
    ],
)
def test_widget_create_invalid_content(value):
    """If the content provided at construction isn't 2- or 3-tuples, an error is
    raised."""
    with pytest.raises(
        ValueError,
        match=(
            r"Content items must be an OptionItem instance, or tuples of \(title, widget\), "
            r"\(title, widget, icon\), or \(title, widget, icon, enabled\)"
        ),
    ):
        toga.OptionContainer(content=value)


def test_item_create(content1):
    """An OptionItem can be created."""
    item = toga.OptionItem("label", content1)

    assert item.text == "label"
    assert item.content == content1
    assert item.index is None
    assert item.interface is None


@pytest.mark.parametrize(
    "title, has_content, has_icon, enabled, message",
    [
        (None, True, True, True, r"Item text cannot be None"),
        ("", True, True, True, r"Item text cannot be blank"),
        ("label", False, True, True, r"Content widget cannot be None"),
    ],
)
def test_item_create_invalid_item(
    title,
    has_content,
    has_icon,
    enabled,
    message,
    content1,
    tab_icon,
):
    """If item details are invalid, an exception is raised."""

    with pytest.raises(ValueError, match=message):
        toga.OptionItem(
            title,
            content1 if has_content else None,
            icon=tab_icon if has_icon else None,
            enabled=enabled,
        )


def test_assign_to_app(app, optioncontainer, content1, content2, content3):
    """If the widget is assigned to an app, the content is also assigned."""
    # Option container is initially unassigned
    assert optioncontainer.app is None

    # Assign the option container to the app
    optioncontainer.app = app

    # option container is on the app
    assert optioncontainer.app == app

    # Content is also on the app
    assert content1.app == app
    assert content2.app == app
    assert content3.app == app


def test_assign_to_app_no_content(app):
    """If the widget is assigned to an app, and there is no content, there's no
    error."""
    optioncontainer = toga.OptionContainer()

    # Option container is initially unassigned
    assert optioncontainer.app is None

    # Assign the Option container to the app
    optioncontainer.app = app

    # Option container is on the app
    assert optioncontainer.app == app


def test_assign_to_window(window, optioncontainer, content1, content2, content3):
    """If the widget is assigned to a window, the content is also assigned."""
    # Option container is initially unassigned
    assert optioncontainer.window is None

    # Assign the Option container to the window
    optioncontainer.window = window

    # Option container is on the window
    assert optioncontainer.window == window
    # Content is also on the window
    assert content1.window == window
    assert content2.window == window
    assert content3.window == window


def test_assign_to_window_no_content(window):
    """If the widget is assigned to a window, and there is no content, there's no
    error."""
    optioncontainer = toga.OptionContainer()

    # Option container is initially unassigned
    assert optioncontainer.window is None

    # Assign the Option container to the window
    optioncontainer.window = window

    # Option container is on the window
    assert optioncontainer.window == window


def test_disable_no_op(optioncontainer):
    """OptionContainer doesn't have a disabled state."""
    # Enabled by default
    assert optioncontainer.enabled

    # Try to disable the widget
    optioncontainer.enabled = False

    # Still enabled.
    assert optioncontainer.enabled


def test_focus_noop(optioncontainer):
    """Focus is a no-op."""

    optioncontainer.focus()
    assert_action_not_performed(optioncontainer, "focus")


@pytest.mark.parametrize("bare_item", [True, False])
@pytest.mark.parametrize(
    "value, expected",
    [
        (None, False),
        ("", False),
        ("true", True),
        ("false", True),  # Evaluated as a string, this value is true.
        (0, False),
        (1234, True),
    ],
)
def test_item_enabled(optioncontainer, value, expected, bare_item):
    """The enabled status of an item can be changed."""
    if bare_item:
        item = toga.OptionItem("title", toga.Box())
    else:
        item = optioncontainer.content[1]

    # item is initially enabled by default.
    assert item.enabled

    # Set the enabled status
    item.enabled = value
    assert item.enabled == expected

    # Disable the widget
    item.enabled = False
    assert not item.enabled

    # Set the enabled status again
    item.enabled = value
    assert item.enabled == expected


def test_disable_current_item(optioncontainer):
    """The currently selected item cannot be disabled."""
    # Item 0 is selected by default
    item = optioncontainer.content[0]
    with pytest.raises(
        ValueError,
        match=r"The currently selected tab cannot be disabled.",
    ):
        item.enabled = False

    # Try disabling the current tab directly
    with pytest.raises(
        ValueError,
        match=r"The currently selected tab cannot be disabled.",
    ):
        optioncontainer.current_tab.enabled = False


class MyTitle:
    def __init__(self, title):
        self.title = title

    def __str__(self):
        return self.title


@pytest.mark.parametrize("bare_item", [True, False])
@pytest.mark.parametrize(
    "value, expected",
    [
        ("New Title", "New Title"),
        (42, "42"),  # Evaluated as a string
        (MyTitle("Custom Title"), "Custom Title"),  # Evaluated as a string
    ],
)
def test_item_text(optioncontainer, value, expected, bare_item):
    """The title of an item can be changed."""
    if bare_item:
        item = toga.OptionItem("title", toga.Box())
    else:
        item = optioncontainer.content[1]

    # Set the item text
    item.text = value
    assert item.text == expected


@pytest.mark.parametrize("bare_item", [True, False])
@pytest.mark.parametrize(
    "value, error",
    [
        (None, r"Item text cannot be None"),
        ("", r"Item text cannot be blank"),
        (MyTitle(""), r"Item text cannot be blank"),
    ],
)
def test_invalid_item_text(optioncontainer, value, error, bare_item):
    """Invalid item titles are prevented."""
    if bare_item:
        item = toga.OptionItem("title", toga.Box())
    else:
        item = optioncontainer.content[1]

    # Using invalid text raises an error
    with pytest.raises(ValueError, match=error):
        item.text = value


@pytest.mark.parametrize("bare_item", [True, False])
def test_item_icon(optioncontainer, bare_item):
    """The icon of an item can be changed."""
    if bare_item:
        item = toga.OptionItem("title", toga.Box())
    else:
        item = optioncontainer.content[0]

    # Icon is initially empty
    assert item.icon is None

    test_icon = toga.Icon("test-icon")
    item.icon = test_icon

    # Icon has been set
    assert item.icon == test_icon
    if bare_item:
        assert_action_not_performed(optioncontainer, "set option icon")
    else:
        assert_action_performed_with(
            optioncontainer,
            "set option icon",
            index=0,
        )
    EventLog.reset()

    # Clear the icon
    item.icon = None

    # Icon has been reset
    assert item.icon is None
    if bare_item:
        assert_action_not_performed(optioncontainer, "set option icon")
    else:
        assert_action_performed_with(
            optioncontainer,
            "set option icon",
            index=0,
            icon=None,
        )
    EventLog.reset()

    # Icon has been set by name
    item.icon = "new-icon"

    # Icon has been set to the new value
    assert item.icon.path.name == "new-icon"
    if bare_item:
        assert_action_not_performed(optioncontainer, "set option icon")
    else:
        assert_action_performed_with(
            optioncontainer,
            "set option icon",
            index=0,
        )


@pytest.mark.parametrize("bare_item", [True, False])
def test_item_icon_disabled(monkeypatch, optioncontainer, bare_item):
    """The icon of an item won't be set if icons aren't in use."""
    # monkeypatch the OptionContainer class to disable the use of icons
    monkeypatch.setattr(dummy_factory.OptionContainer, "uses_icons", False)

    if bare_item:
        item = toga.OptionItem("title", toga.Box())
    else:
        item = optioncontainer.content[0]

    # Icon is initially empty
    assert item.icon is None

    # Try to set an icon
    item.icon = "test-icon"

    # Icon is still none
    assert item.icon is None

    # Add a new content item with an icon
    optioncontainer.content.append("New content", toga.Box(), icon="new-icon")

    # Icon is still none
    assert optioncontainer.content["New content"].icon is None


def test_optionlist_repr(optioncontainer):
    """OptionContainer content has a helpful repr."""
    assert (
        repr(optioncontainer.content)
        == "<OptionList 'Item 1', 'Item 2', 'Item 3', 'Item 4'>"
    )


def test_optionlist_iter(optioncontainer):
    """OptionContainer content can be iterated."""
    assert [item.text for item in optioncontainer.content] == [
        "Item 1",
        "Item 2",
        "Item 3",
        "Item 4",
    ]


def test_optionlist_len(optioncontainer):
    """OptionContainer content has length."""
    assert len(optioncontainer.content) == 4


@pytest.mark.parametrize("index", [1, "Item 2", None])
def test_getitem(optioncontainer, content2, index):
    """An item can be retrieved."""
    if index is None:
        index = optioncontainer.content[1]

    # get item
    item = optioncontainer.content[index]
    assert item.text == "Item 2"
    assert item.index == 1
    assert item.content == content2


@pytest.mark.parametrize("index", [1, "Item 2", None])
def test_delitem(optioncontainer, index):
    """An item can be removed with __del__"""
    if index is None:
        index = optioncontainer.content[1]

    # get a reference to items 1 and 3
    item1 = optioncontainer.content[0]
    item3 = optioncontainer.content[2]

    # delete item
    del optioncontainer.content[index]
    assert len(optioncontainer.content) == 3
    assert_action_performed_with(optioncontainer, "remove option", index=1)

    # There's no item with the deleted label
    with pytest.raises(ValueError, match=r"No tab named 'Item 2'"):
        optioncontainer.content.index("Item 2")

    # The index of item 3 has been reduced; item 1 is untouched
    assert item1.index == 0
    assert item3.index == 1

    # Widget has been refreshed
    assert_action_performed(optioncontainer, "refresh")


@pytest.mark.parametrize("index", [0, "Item 1", None])
def test_delitem_current(optioncontainer, index):
    """The current item can't be deleted."""
    if index is None:
        index = optioncontainer.content[0]

    with pytest.raises(
        ValueError, match=r"The currently selected tab cannot be deleted."
    ):
        del optioncontainer.content[index]


@pytest.mark.parametrize("index", [1, "Item 2", None])
def test_item_remove(optioncontainer, index):
    """An item can be removed with remove."""
    if index is None:
        index = optioncontainer.content[1]

    # get a reference to items 1 and 3
    item1 = optioncontainer.content[0]
    item3 = optioncontainer.content[2]

    # remove item
    optioncontainer.content.remove(index)
    assert len(optioncontainer.content) == 3
    assert_action_performed_with(optioncontainer, "remove option", index=1)

    # There's no item with the deleted label
    with pytest.raises(ValueError, match=r"No tab named 'Item 2'"):
        optioncontainer.content.index("Item 2")

    # The index of item 3 has been reduced; item 1 is untouched
    assert item1.index == 0
    assert item3.index == 1

    # Widget has been refreshed
    assert_action_performed(optioncontainer, "refresh")


@pytest.mark.parametrize("index", [0, "Item 1", None])
def test_item_remove_current(optioncontainer, index):
    """The current item can't be removed."""
    if index is None:
        index = optioncontainer.content[0]

    with pytest.raises(
        ValueError, match=r"The currently selected tab cannot be deleted."
    ):
        optioncontainer.content.remove(index)


def test_item_insert_item(optioncontainer):
    """The text of an inserted item can be set."""
    new_content = toga.Box()
    item = toga.OptionItem("New Tab", new_content)

    optioncontainer.content.insert(1, item)

    # Backend added an item and set enabled
    assert_action_performed_with(
        optioncontainer,
        "add option",
        index=1,
        text="New Tab",
        widget=new_content._impl,
    )
    assert_action_performed_with(
        optioncontainer,
        "set option enabled",
        index=1,
        value=True,
    )
    assert_action_performed_with(optioncontainer, "refresh")


@pytest.mark.parametrize(
    "args,kwargs,message",
    [
        (
            ("New Tab",),
            {},
            r"Content widget cannot be None.",
        ),
        (
            (toga.OptionItem("New Tab", toga.Box()),),
            {"content": toga.Box()},
            r"Cannot specify content if using an OptionItem instance.",
        ),
        (
            (toga.OptionItem("New Tab", toga.Box()),),
            {"icon": "tab-icon"},
            r"Cannot specify icon if using an OptionItem instance.",
        ),
        (
            (toga.OptionItem("New Tab", toga.Box()),),
            {"enabled": False},
            r"Cannot specify enabled if using an OptionItem instance.",
        ),
    ],
)
def test_item_insert_item_invalid(optioncontainer, args, kwargs, message):
    """If both an item and specific details are provided, an error is raised."""
    with pytest.raises(ValueError, match=message):
        optioncontainer.content.insert(1, *args, **kwargs)


@pytest.mark.parametrize(
    "value, expected",
    [
        ("New Title", "New Title"),
        (42, "42"),  # Evaluated as a string
        (MyTitle("Custom Title"), "Custom Title"),  # Evaluated as a string
    ],
)
def test_item_insert_text(optioncontainer, value, expected):
    """The text of an inserted item can be set."""
    new_content = toga.Box()

    optioncontainer.content.insert(1, value, new_content, enabled=True)

    # Backend added an item and set enabled
    assert_action_performed_with(
        optioncontainer,
        "add option",
        index=1,
        text=expected,
        widget=new_content._impl,
    )
    assert_action_performed_with(
        optioncontainer,
        "set option enabled",
        index=1,
        value=True,
    )
    assert_action_performed_with(optioncontainer, "refresh")


@pytest.mark.parametrize(
    "value, error",
    [
        (None, r"Item text cannot be None"),
        ("", r"Item text cannot be blank"),
        (MyTitle(""), r"Item text cannot be blank"),
    ],
)
def test_item_insert_invalid_text(optioncontainer, value, error):
    """The item text must be valid."""
    new_content = toga.Box()
    with pytest.raises(ValueError, match=error):
        optioncontainer.content.insert(1, value, new_content, enabled=True)


@pytest.mark.parametrize("enabled", [True, False])
def test_item_insert_enabled(optioncontainer, enabled):
    """The enabled status of content can be set on insert."""
    new_content = toga.Box()

    optioncontainer.content.insert(1, "New content", new_content, enabled=enabled)

    # Backend added an item and set enabled
    assert_action_performed_with(
        optioncontainer,
        "add option",
        index=1,
        text="New content",
        widget=new_content._impl,
    )
    assert_action_performed_with(
        optioncontainer,
        "set option enabled",
        index=1,
        value=enabled,
    )
    assert_action_performed_with(optioncontainer, "refresh")


@pytest.mark.parametrize("enabled", [True, False])
def test_item_append(optioncontainer, enabled):
    """An item can be appended to the content list."""
    # append is implemented using insert;
    # the bulk of the functionality is tested there.
    new_content = toga.Box()

    optioncontainer.content.append("New content", new_content, enabled=enabled)
    assert_action_performed_with(
        optioncontainer, "add option", index=4, widget=new_content._impl
    )
    assert_action_performed_with(
        optioncontainer, "set option enabled", index=4, value=enabled
    )
    assert_action_performed_with(optioncontainer, "refresh")


@pytest.mark.parametrize("index", [1, "Item 2", None])
def test_current_tab(optioncontainer, index, on_select_handler):
    """The current tab of the optioncontainer can be changed."""
    if index is None:
        index = optioncontainer.content[1]

    # First item is selected initially
    assert optioncontainer.current_tab.index == 0
    assert optioncontainer.current_tab.text == "Item 1"

    # Programmatically select item 2
    optioncontainer.current_tab = index

    # Current tab values have changed
    assert optioncontainer.current_tab.index == 1
    assert optioncontainer.current_tab.text == "Item 2"

    # on_select handler was invoked
    on_select_handler.assert_called_once_with(optioncontainer)


def test_select_disabled_tab(optioncontainer):
    """A disabled tab cannot be selected."""

    # Disable item 1
    item = optioncontainer.content[1]
    item.enabled = False

    with pytest.raises(
        ValueError,
        match=r"A disabled tab cannot be made the current tab.",
    ):
        optioncontainer.current_tab = 1
